@@ -62,15 +62,15 @@ buffer_init (Buffer *buffer)
 
 /**
  * gnome_vfs_socket_buffer_new:
- * @socket: socket to be buffered
+ * @socket: socket to be buffered.
  *
  * Create a socket buffer around @socket. A buffered
  * socket allows data to be poked at without reading it
  * as it will be buffered. A future read will retrieve
  * the data again.
  *
- * Return value: a newly allocated GnomeVFSSocketBuffer
- **/
+ * Return value: a newly allocated #GnomeVFSSocketBuffer.
+ */
 GnomeVFSSocketBuffer*  
 gnome_vfs_socket_buffer_new (GnomeVFSSocket *socket)
 {
@@ -89,13 +89,14 @@ gnome_vfs_socket_buffer_new (GnomeVFSSocket *socket)
 
 /**
  * gnome_vfs_socket_buffer_destroy:
- * @socket_buffer: buffered socket to destray
- * @close_socket: if %TRUE the socket being buffered will be closed too
+ * @socket_buffer: buffered socket to destroy.
+ * @close_socket: if %TRUE, the socket being buffered will be closed.
+ * @cancellation: handle allowing cancellation of the operation.
  *
  * Free the socket buffer.
  *
- * Return value: GnomeVFSResult indicating the success of the operation
- **/
+ * Return value: #GnomeVFSResult indicating the success of the operation.
+ */
 GnomeVFSResult   
 gnome_vfs_socket_buffer_destroy  (GnomeVFSSocketBuffer *socket_buffer, 
 				  gboolean close_socket,
@@ -120,44 +121,52 @@ refill_input_buffer (GnomeVFSSocketBuffer *socket_buffer,
 	Buffer *input_buffer;
 	GnomeVFSResult result;
 	GnomeVFSFileSize bytes_read;
-	
+	char *data_pos;
+
 	input_buffer = &socket_buffer->input_buffer;
 
-	if (input_buffer->last_error != GNOME_VFS_OK
-	    || input_buffer->byte_count > 0) {
+	if (input_buffer->last_error != GNOME_VFS_OK) {
 		return FALSE;
 	}
 
-	input_buffer->offset = 0;
 
+	/* If there is data left in the buffer move it to the front */
+	if (input_buffer->offset > 0) {
+		data_pos = &(input_buffer->data[input_buffer->offset]);
+		memmove (input_buffer->data, data_pos, input_buffer->byte_count);
+		input_buffer->offset = 0;
+	}
+	
 	result = gnome_vfs_socket_read (socket_buffer->socket,
-					&input_buffer->data, 
-					BUFFER_SIZE,
+					input_buffer->data + input_buffer->byte_count,
+					BUFFER_SIZE - input_buffer->byte_count,
 					&bytes_read,
 					cancellation);
+
 	
 	if (result != GNOME_VFS_OK) {
 		input_buffer->last_error = result;
 		return FALSE;
 	}
 
-	input_buffer->byte_count = bytes_read;
+	input_buffer->byte_count += bytes_read;
 
 	return TRUE;
 }
 
 /**
  * gnome_vfs_socket_buffer_read:
- * @socket_buffer: buffered socket to read data from
- * @buffer: allocated buffer of at least @bytes bytes to be read into
- * @bytes: number of bytes to read from @socket into @socket_buffer
- * @bytes_read: pointer to a GnomeVFSFileSize, will contain
- * the number of bytes actually read from the socket on return.
+ * @socket_buffer: buffered socket to read data from.
+ * @buffer: allocated buffer of at least @bytes bytes to be read into.
+ * @bytes: number of bytes to read from @socket_buffer into @buffer.
+ * @bytes_read: pointer to a #GnomeVFSFileSize, will contain
+ * the number of bytes actually read from the @socket_buffer on return.
+ * @cancellation: handle allowing cancellation of the operation.
  *
  * Read @bytes bytes of data from the @socket into @socket_buffer.
  *
- * Return value: GnomeVFSResult indicating the success of the operation
- **/
+ * Return value: #GnomeVFSResult indicating the success of the operation.
+ */
 GnomeVFSResult   
 gnome_vfs_socket_buffer_read (GnomeVFSSocketBuffer *socket_buffer,
 			      gpointer buffer,
@@ -176,7 +185,9 @@ gnome_vfs_socket_buffer_read (GnomeVFSSocketBuffer *socket_buffer,
 	   "If nbyte is 0, read() will return 0 and have no other results."
 	*/
 	if (bytes == 0) {
-		*bytes_read = 0;
+		if (bytes_read != NULL);
+			*bytes_read = 0;
+
 		return GNOME_VFS_OK;
 	}
 		
@@ -193,35 +204,160 @@ gnome_vfs_socket_buffer_read (GnomeVFSSocketBuffer *socket_buffer,
 		}
 	}
 
+	n = 0;
+	
 	if (input_buffer->byte_count != 0) {
 		n = MIN (bytes, input_buffer->byte_count);
 		memcpy (buffer, input_buffer->data + input_buffer->offset, n);
 		input_buffer->byte_count -= n;
 		input_buffer->offset += n;
-		if (bytes_read != NULL) {
-			*bytes_read = n;
-		}
-	} else {
-		if (bytes_read != NULL) {
+	}
+	
+	if (bytes_read != NULL) {
+		*bytes_read = n;
+	}
+	
+	return result;
+}
+
+/**
+ * gnome_vfs_socket_buffer_read_until:
+ * @socket_buffer: buffered socket to read data from.
+ * @buffer: allocated buffer of at least @bytes bytes to be read into.
+ * @bytes: maximum number of bytes to read from @socket_buffer into @buffer.
+ * @boundary: the boundary until which is read.
+ * @boundary_len: the length of the @boundary.
+ * @bytes_read: pointer to a #GnomeVFSFileSize, will contain
+ * the number of bytes actually read from the @socket_buffer on return.
+ * @got_boundary: pointer to a #gboolean  which will be %TRUE if the boundary
+ * was found or %FALSE otherwise.
+ * @cancellation: handle allowing cancellation of the operation.
+ *
+ * Read up to @bytes bytes of data from the @socket_buffer into @buffer 
+ * until boundary is reached. @got_boundary will be set accordingly.
+ *
+ * Note that if @bytes is smaller than @boundary_len there is no way
+ * to detected the boundary! So if you want to make sure that every boundary
+ * is found (in a loop maybe) assure that @bytes is at least as big as 
+ * @boundary_len.
+ *
+ * Return value: #GnomeVFSResult indicating the success of the operation.
+ *
+ * Since: 2.8
+ */
+GnomeVFSResult
+gnome_vfs_socket_buffer_read_until (GnomeVFSSocketBuffer *socket_buffer,
+				    gpointer buffer,
+				    GnomeVFSFileSize bytes,
+				    gconstpointer boundary,
+				    GnomeVFSFileSize boundary_len,
+				    GnomeVFSFileSize *bytes_read,
+				    gboolean *got_boundary,
+				    GnomeVFSCancellation *cancellation)
+{
+	Buffer *input_buffer;
+	GnomeVFSResult result;
+	GnomeVFSFileSize n, max_scan;
+	char *iter, *start, *delim;
+
+	g_return_val_if_fail (socket_buffer != NULL, GNOME_VFS_ERROR_BAD_PARAMETERS);
+	g_return_val_if_fail (buffer != NULL, GNOME_VFS_ERROR_BAD_PARAMETERS);
+	g_return_val_if_fail (boundary != NULL, GNOME_VFS_ERROR_BAD_PARAMETERS);
+	g_return_val_if_fail (got_boundary != NULL, GNOME_VFS_ERROR_BAD_PARAMETERS);
+	g_return_val_if_fail (boundary_len < BUFFER_SIZE, GNOME_VFS_ERROR_TOO_BIG);
+
+	*got_boundary = FALSE;
+
+	/* Quote from UNIX 98:
+	   "If nbyte is 0, read() will return 0 and have no other results."
+	*/
+	if (bytes == 0) {
+		if (bytes_read != NULL)
 			*bytes_read = 0;
+
+		return GNOME_VFS_OK;
+	}
+
+	input_buffer = &socket_buffer->input_buffer;
+	result = GNOME_VFS_OK;
+	
+	/* we are looping here to catch the case where we are close
+	 * to eof and haveing less or equal bytes then boundary_len */
+	while (input_buffer->byte_count <= boundary_len) {
+		if (! refill_input_buffer (socket_buffer, cancellation)) {
+			break;
 		}
 	}
 
+	/* At this point we have either byte_count > boundary_len or
+	 * we have and error during refill */
+
+	n = 0;
+	start = input_buffer->data + input_buffer->offset;
+	max_scan = MIN (input_buffer->byte_count, bytes);
+
+	/* if max_scan is greater then boundary_len do a scan (I) 
+	 * otherwise we had an error during the loop above or bytes is
+	 * too small. (II) We handle the case where boundary_len ==
+	 * max_scan in (II). */
+
+	if (max_scan > boundary_len) {
+
+		delim = start + max_scan;
+		for (iter = start; iter + boundary_len <= delim; iter++) {
+			if (!memcmp (iter, boundary, boundary_len)) {
+				*got_boundary = TRUE;
+				/* We wanna have the boundary fetched */
+				iter += boundary_len;
+				break;
+			}
+		}
+
+		/* Fetch data data until iter */
+		n = iter - start;
+
+	} else /* (II) */ {
+
+		if (max_scan == boundary_len &&
+			!memcmp (start, boundary, boundary_len)) {
+			*got_boundary = TRUE;
+		}
+
+		n = max_scan;
+	}
+
+	if (n > 0) {
+
+		memcpy (buffer, start, n);
+		input_buffer->byte_count -= n;
+		input_buffer->offset += n;
+		/* queque up the fill buffer error if any
+		 * until the buffer is flushed */
+	} else {
+		result = input_buffer->last_error;
+		input_buffer->last_error = GNOME_VFS_OK;
+	}
+
+	if (bytes_read != NULL) {
+		*bytes_read = n;
+	}
+
 	return result;
 }
 
 /**
  * gnome_vfs_socket_buffer_peekc:
- * @socket_buffer: the socket buffer to read from
+ * @socket_buffer: the socket buffer to read from.
  * @character: pointer to a char, will contain a character on return from
- * a successful "peek"
+ * a successful "peek".
+ * @cancellation: handle allowing cancellation of the operation.
  *
  * Peek at the next character in @socket_buffer without actually reading
  * the character in. The next read will retrieve @c (as well as any following
  * data if requested).
  *
- * Return value: GnomeVFSResult indicating the success of the operation
- **/
+ * Return value: #GnomeVFSResult indicating the success of the operation.
+ */
 GnomeVFSResult
 gnome_vfs_socket_buffer_peekc (GnomeVFSSocketBuffer *socket_buffer,
 			       gchar *character,
@@ -275,7 +411,10 @@ flush (GnomeVFSSocketBuffer *socket_buffer,
 		if (result != GNOME_VFS_OK) {
 			return result;
 		}
-		
+
+		memmove (output_buffer->data,
+			 output_buffer->data + bytes_written,
+			 output_buffer->byte_count - bytes_written);
 		output_buffer->byte_count -= bytes_written;
 	}
 
@@ -284,16 +423,17 @@ flush (GnomeVFSSocketBuffer *socket_buffer,
 
 /**
  * gnome_vfs_socket_buffer_write:
- * @socket_buffer: buffered socket to write data to
- * @buffer: data to write to the socket
- * @bytes: number of bytes from @buffer to write to @socket_buffer
- * @bytes_written: pointer to a GnomeVFSFileSize, will contain
- * the number of bytes actually written to the socket on return.
+ * @socket_buffer: buffered socket to write data to.
+ * @buffer: data to write to the @socket_buffer.
+ * @bytes: number of bytes to write from @buffer to @socket_buffer.
+ * @bytes_written: pointer to a #GnomeVFSFileSize, will contain
+ * the number of bytes actually written to the @socket_buffer on return.
+ * @cancellation: handle allowing cancellation of the operation.
  *
  * Write @bytes bytes of data from @buffer to @socket_buffer.
  *
- * Return value: GnomeVFSResult indicating the success of the operation
- **/ 
+ * Return value: #GnomeVFSResult indicating the success of the operation.
+ */ 
 GnomeVFSResult   
 gnome_vfs_socket_buffer_write (GnomeVFSSocketBuffer *socket_buffer, 
 			       gconstpointer buffer,
@@ -345,12 +485,13 @@ gnome_vfs_socket_buffer_write (GnomeVFSSocketBuffer *socket_buffer,
 
 /**
  * gnome_vfs_socket_buffer_flush:
- * @socket_buffer: buffer to flush
+ * @socket_buffer: buffer to flush.
+ * @cancellation: handle allowing cancellation of the operation.
  *
  * Write all outstanding data to @socket_buffer.
  *
- * Return value: GnomeVFSResult indicating the success of the operation
- **/
+ * Return value: #GnomeVFSResult indicating the success of the operation.
+ */
 GnomeVFSResult   
 gnome_vfs_socket_buffer_flush (GnomeVFSSocketBuffer *socket_buffer,
 			       GnomeVFSCancellation *cancellation)
